#![cfg(not(all(target_arch = "wasm32", target_os = "unknown")))]

use cfg_if::*;
use clap::Parser;
use parking_lot::*;
use std::path::PathBuf;
use stop_token::StopSource;
use veilid_tools::*;

const VERSION: &str = env!("CARGO_PKG_VERSION");

cfg_if! {
    if #[cfg(feature="rt-async-std")] {
        pub fn block_on<F: Future<Output = T>, T>(f: F) -> T {
            async_std::task::block_on(f)
        }
    } else if #[cfg(feature="rt-tokio")] {
        pub fn block_on<F: Future<Output = T>, T>(f: F) -> T {
            let rt = tokio::runtime::Runtime::new().unwrap();
            let local = tokio::task::LocalSet::new();
            local.block_on(&rt, f)
        }
    } else {
        compile_error!("needs executor implementation");
    }
}

#[derive(Parser, Debug)]
#[command(author, version, about = "Veilid VirtualRouter")]
struct CmdlineArgs {
    /// TCP address to listen on
    #[arg(short('t'), long)]
    tcp_addr: Option<SocketAddr>,
    /// Turn off TCP listener
    #[arg(long)]
    no_tcp: bool,
    /// WS address to listen on
    #[arg(short('w'), long)]
    ws_addr: Option<SocketAddr>,
    /// Turn off WS listener
    #[arg(long)]
    no_ws: bool,
    /// Specify a configuration file to use
    #[arg(short = 'c', long, value_name = "FILE")]
    config_file: Option<PathBuf>,
    /// Instead of running the virtual router, print the configuration it would use to the console
    #[arg(long)]
    dump_config: bool,
}

fn main() -> Result<(), String> {
    let stop_source = StopSource::new();
    let stop_token = stop_source.token();
    let stop_mutex = Mutex::new(Some(stop_source));

    ctrlc::set_handler(move || {
        println!("Exiting...");
        *(stop_mutex.lock()) = None;
    })
    .expect("Error setting Ctrl-C handler");

    block_on(async {
        println!("Veilid VirtualRouter v{}", VERSION);

        let args = CmdlineArgs::parse();

        let config = config::Config::new(args.config_file)
            .map_err(|e| format!("Error loading config: {}", e))?;

        if args.dump_config {
            let cfg_yaml = serde_yaml::to_string(&config)
                .map_err(|e| format!("Error serializing config: {}", e))?;
            println!("{}", cfg_yaml);
            return Ok(());
        }

        let router_server = virtual_network::RouterServer::new(config);
        let _ss_tcp = if !args.no_tcp {
            Some(
                router_server
                    .listen_tcp(args.tcp_addr)
                    .await
                    .map_err(|e| e.to_string())?,
            )
        } else {
            None
        };

        let _ss_ws = if !args.no_ws {
            Some(
                router_server
                    .listen_ws(args.ws_addr)
                    .await
                    .map_err(|e| e.to_string())?,
            )
        } else {
            None
        };

        router_server
            .run(stop_token)
            .await
            .map_err(|e| e.to_string())
    })
}
