use super::*;
use ipnet::*;
use serde::*;
use std::path::Path;

use validator::{Validate, ValidateArgs, ValidationError, ValidationErrors};

const PREDEFINED_CONFIG: &str = include_str!("predefined_config.yml");
const DEFAULT_CONFIG: &str = include_str!("default_config.yml");

#[derive(Debug, ThisError)]
pub enum ConfigError {
    #[error("parse error")]
    ParseError(::config::ConfigError),
    #[error("validate error")]
    ValidateError(validator::ValidationErrors),
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(context = "ValidateContext<'v_a>")]
pub struct Profile {
    #[validate(
        length(min = 1),
        custom(function = "validate_instances_exist", use_context)
    )]
    pub instances: Vec<Instance>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
#[serde(untagged)]
pub enum Instance {
    Machine { machine: WeightedList<String> },
    Template { template: WeightedList<String> },
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(
    context = "ValidateContext<'v_a>",
    schema(function = "validate_machine", use_context)
)]
pub struct Machine {
    #[serde(flatten)]
    #[validate(custom(function = "validate_machine_location", use_context))]
    pub location: MachineLocation,
    #[serde(default)]
    pub disable_capabilities: Vec<String>,
    #[serde(default)]
    pub bootstrap: bool,
}

fn validate_machine(machine: &Machine, _context: &ValidateContext) -> Result<(), ValidationError> {
    if machine.disable_capabilities.contains(&("".to_string())) {
        return Err(ValidationError::new("badcap")
            .with_message("machine has empty disabled capability".into()));
    }
    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize)]
#[serde(untagged)]
pub enum MachineLocation {
    Network {
        network: String,
        #[serde(default)]
        address4: Option<Ipv4Addr>,
        #[serde(default)]
        address6: Option<Ipv6Addr>,
    },
    Blueprint {
        blueprint: String,
    },
}

fn validate_machine_location(
    value: &MachineLocation,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    match value {
        MachineLocation::Network {
            network,
            address4,
            address6,
        } => {
            if address4.is_none() && address6.is_none() {
                return Err(ValidationError::new("badaddr")
                    .with_message("machine must have at least one address".into()));
            }
            validate_network_exists(network, context)?;
        }
        MachineLocation::Blueprint { blueprint } => {
            validate_blueprint_exists(blueprint, context)?;
        }
    }

    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(
    context = "ValidateContext<'v_a>",
    schema(function = "validate_template", use_context)
)]
pub struct Template {
    #[serde(flatten)]
    #[validate(custom(function = "validate_template_location", use_context))]
    pub location: TemplateLocation,
    #[serde(flatten)]
    #[validate(nested)]
    pub limits: TemplateLimits,
    #[serde(default)]
    pub disable_capabilities: Vec<String>,
}

fn validate_template(
    template: &Template,
    _context: &ValidateContext,
) -> Result<(), ValidationError> {
    if template.disable_capabilities.contains(&("".to_string())) {
        return Err(ValidationError::new("badcap")
            .with_message("template has empty disabled capability".into()));
    }
    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(schema(function = "validate_template_limits"))]
pub struct TemplateLimits {
    /// maximum number of machines this template will generate
    #[validate(nested)]
    #[serde(default)]
    pub machine_count: Option<WeightedList<u32>>,
    #[validate(nested)]
    pub machines_per_network: Option<WeightedList<u32>>,
}

fn validate_template_limits(limits: &TemplateLimits) -> Result<(), ValidationError> {
    if let Some(machine_count) = &limits.machine_count {
        machine_count.try_for_each(|x| {
            if *x == 0 {
                return Err(ValidationError::new("badcount")
                    .with_message("template limits has zero machine count".into()));
            }
            Ok(())
        })?;
    }
    if let Some(machines_per_network) = &limits.machines_per_network {
        machines_per_network.try_for_each(|x| {
            if *x == 0 {
                return Err(ValidationError::new("badcount")
                    .with_message("template limits has zero machines per network count".into()));
            }
            Ok(())
        })?;
    }

    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize)]
#[serde(untagged)]
pub enum TemplateLocation {
    Network { network: WeightedList<String> },
    Blueprint { blueprint: WeightedList<String> },
}

fn validate_template_location(
    value: &TemplateLocation,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    match value {
        TemplateLocation::Network { network } => {
            network.try_for_each(|m| validate_network_exists(m, context))?;
        }
        TemplateLocation::Blueprint { blueprint } => {
            blueprint.try_for_each(|t| validate_blueprint_exists(t, context))?;
        }
    }

    Ok(())
}

////////////////////////////////////////////////////////////////

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(
    context = "ValidateContext<'v_a>",
    schema(function = "validate_network", use_context)
)]
pub struct Network {
    #[serde(default)]
    #[validate(custom(function = "validate_model_exists", use_context))]
    pub model: Option<String>,
    #[serde(default)]
    #[validate(custom(function = "validate_network_ipv4", use_context))]
    pub ipv4: Option<NetworkIpv4>,
    #[serde(default)]
    #[validate(custom(function = "validate_network_ipv6", use_context))]
    pub ipv6: Option<NetworkIpv6>,
}

fn validate_network(network: &Network, _context: &ValidateContext) -> Result<(), ValidationError> {
    if network.ipv4.is_none() && network.ipv6.is_none() {
        return Err(ValidationError::new("badaddr")
            .with_message("network must support at least one address type".into()));
    }
    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct NetworkIpv4 {
    pub allocation: String,
    #[serde(default)]
    pub gateway: Option<NetworkGateway>,
}

fn validate_network_ipv4(
    network_ipv4: &NetworkIpv4,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    validate_allocation_exists(&network_ipv4.allocation, context)?;
    if let Some(gateway) = &network_ipv4.gateway {
        validate_network_gateway(gateway, context)?;
    }
    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct NetworkIpv6 {
    pub allocation: String,
    #[serde(default)]
    pub gateway: Option<NetworkGateway>,
}

fn validate_network_ipv6(
    network_ipv6: &NetworkIpv6,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    validate_allocation_exists(&network_ipv6.allocation, context)?;
    if let Some(gateway) = &network_ipv6.gateway {
        validate_network_gateway(gateway, context)?;
    }
    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
pub struct NetworkGateway {
    pub translation: Translation,
    pub upnp: bool,
    pub network: Option<String>,
}

fn validate_network_gateway(
    gateway: &NetworkGateway,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    if let Some(network) = &gateway.network {
        validate_network_exists(network, context)?;
    }
    Ok(())
}
////////////////////////////////////////////////////////////////

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(
    context = "ValidateContext<'v_a>",
    schema(function = "validate_blueprint", use_context)
)]
pub struct Blueprint {
    #[serde(default)]
    #[validate(custom(function = "validate_models_exist", use_context))]
    pub model: Option<WeightedList<String>>,
    #[validate(nested)]
    pub limits: BlueprintLimits,
    #[serde(default)]
    #[validate(custom(function = "validate_blueprint_ipv4", use_context))]
    pub ipv4: Option<BlueprintIpv4>,
    #[serde(default)]
    #[validate(custom(function = "validate_blueprint_ipv6", use_context))]
    pub ipv6: Option<BlueprintIpv6>,
}

fn validate_blueprint(
    blueprint: &Blueprint,
    _context: &ValidateContext,
) -> Result<(), ValidationError> {
    if blueprint.ipv4.is_none() && blueprint.ipv6.is_none() {
        return Err(ValidationError::new("badaddr")
            .with_message("blueprint must support at least one address type".into()));
    }
    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(schema(function = "validate_blueprint_limits"))]
pub struct BlueprintLimits {
    /// maximum number of networks this blueprint will generate
    #[validate(nested)]
    #[serde(default)]
    pub network_count: Option<WeightedList<u32>>,
}

fn validate_blueprint_limits(limits: &BlueprintLimits) -> Result<(), ValidationError> {
    if let Some(network_count) = &limits.network_count {
        network_count.try_for_each(|x| {
            if *x == 0 {
                return Err(ValidationError::new("badcount")
                    .with_message("blueprint limits has zero network count".into()));
            }
            Ok(())
        })?;
    }

    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize)]
#[serde(untagged)]
pub enum BlueprintLocation {
    Allocation {
        allocation: WeightedList<String>,
    },
    Network {
        #[serde(default)]
        network: Option<WeightedList<String>>,
    },
}

fn validate_blueprint_location(
    value: &BlueprintLocation,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    match value {
        BlueprintLocation::Allocation { allocation } => {
            allocation.try_for_each(|a| validate_allocation_exists(a, context))?;
        }
        BlueprintLocation::Network { network } => {
            if let Some(network) = network {
                network.try_for_each(|n| validate_network_exists(n, context))?;
            }
        }
    }

    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct BlueprintIpv4 {
    #[serde(flatten)]
    pub location: BlueprintLocation,
    pub prefix: WeightedList<u8>,
    #[serde(default)]
    pub gateway: Option<BlueprintGateway>,
}

fn validate_blueprint_ipv4(
    blueprint_ipv4: &BlueprintIpv4,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    validate_blueprint_location(&blueprint_ipv4.location, context)?;
    blueprint_ipv4.prefix.validate_once()?;
    blueprint_ipv4.prefix.try_for_each(|x| {
        if *x > 32 {
            return Err(ValidationError::new("badprefix")
                .with_message("ipv4 blueprint prefix too long".into()));
        }
        Ok(())
    })?;

    if let Some(gateway) = &blueprint_ipv4.gateway {
        validate_blueprint_gateway(gateway, context)?;
    }
    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct BlueprintIpv6 {
    #[serde(flatten)]
    pub allocation: BlueprintLocation,
    pub prefix: WeightedList<u8>,
    #[serde(default)]
    pub gateway: Option<BlueprintGateway>,
}

fn validate_blueprint_ipv6(
    blueprint_ipv6: &BlueprintIpv6,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    validate_blueprint_location(&blueprint_ipv6.allocation, context)?;
    blueprint_ipv6.prefix.validate_once()?;
    blueprint_ipv6.prefix.try_for_each(|x| {
        if *x > 128 {
            return Err(ValidationError::new("badprefix")
                .with_message("ipv6 blueprint prefix too long".into()));
        }
        Ok(())
    })?;

    if let Some(gateway) = &blueprint_ipv6.gateway {
        validate_blueprint_gateway(gateway, context)?;
    }
    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct BlueprintGateway {
    pub translation: WeightedList<Translation>,
    pub upnp: Probability,
    #[serde(flatten)]
    pub location: TemplateLocation,
}

fn validate_blueprint_gateway(
    gateway: &BlueprintGateway,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    gateway.translation.validate_once()?;
    validate_template_location(&gateway.location, context)?;
    Ok(())
}
////////////////////////////////////////////////////////////////

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(schema(function = "validate_subnets"))]
pub struct Subnets {
    #[serde(default)]
    pub subnet4: Option<WeightedList<Ipv4Net>>,
    #[serde(default)]
    pub subnet6: Option<WeightedList<Ipv6Net>>,
}

fn validate_subnets(subnets: &Subnets) -> Result<(), ValidationError> {
    if subnets.subnet4.is_none() && subnets.subnet6.is_none() {
        return Err(ValidationError::new("badsub")
            .with_message("subnets must support at least one address type".into()));
    }
    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(schema(function = "validate_distance"))]
pub struct Distance {
    pub min: f32,
    pub max: f32,
}

fn validate_distance(distance: &Distance) -> Result<(), ValidationError> {
    if distance.min < 0.0 {
        return Err(ValidationError::new("baddist")
            .with_message("distance minimum must not be negative".into()));
    }
    if distance.max < distance.min {
        return Err(ValidationError::new("baddist")
            .with_message("distance maximum must not be less than the minimum".into()));
    }
    Ok(())
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(schema(function = "validate_distribution"))]
pub struct Distribution {
    pub mean: f32,
    pub sigma: f32,
    pub skew: f32,
    pub min: f32,
    pub max: f32,
}

fn validate_distribution(distribution: &Distribution) -> Result<(), ValidationError> {
    if distribution.mean < 0.0 {
        return Err(ValidationError::new("baddistrib")
            .with_message("distribution mean must not be negative".into()));
    }
    if distribution.sigma < distribution.mean {
        return Err(ValidationError::new("baddistrib")
            .with_message("distribution sigma must not be less than the mean".into()));
    }
    if distribution.max < distribution.min {
        return Err(ValidationError::new("baddistrib")
            .with_message("distribution maximum must not be less than the minimum".into()));
    }
    Ok(())
}

#[derive(Debug, Copy, Clone, Serialize, Deserialize)]
#[serde(rename_all = "snake_case")]
pub enum Translation {
    None,
    PortRestricted,
    AddressRestricted,
    Symmetric,
}

impl Default for Translation {
    fn default() -> Self {
        Self::None
    }
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(context = "ValidateContext<'v_a>")]
pub struct Model {
    #[validate(nested)]
    pub latency: Distribution,
    #[serde(default)]
    #[validate(nested)]
    pub distance: Option<Distance>,
    #[serde(default)]
    #[validate(range(min = 0.0, max = 1.0))]
    pub loss: Probability,
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(context = "ValidateContext<'v_a>")]
pub struct Allocation {
    #[serde(flatten)]
    #[validate(nested)]
    pub subnets: Subnets,
}

pub struct ValidateContext<'a> {
    config: &'a Config,
}

#[derive(Debug, Clone, Serialize, Deserialize, Validate)]
#[validate(context = "ValidateContext<'v_a>")]
pub struct Config {
    pub seed: Option<u64>,
    #[validate(
        length(min = 1),
        custom(function = "validate_network_exists", use_context)
    )]
    pub default_network: String,
    #[validate(
        length(min = 1),
        custom(function = "validate_model_exists", use_context)
    )]
    pub default_model: String,
    #[serde(default)]
    #[validate(length(min = 1))]
    pub profiles: HashMap<String, Profile>,
    #[serde(default)]
    pub machines: HashMap<String, Machine>,
    #[serde(default)]
    pub templates: HashMap<String, Template>,
    #[serde(default)]
    pub networks: HashMap<String, Network>,
    #[serde(default)]
    pub blueprints: HashMap<String, Blueprint>,
    #[serde(default)]
    pub allocations: HashMap<String, Allocation>,
    #[serde(default)]
    pub models: HashMap<String, Model>,
}

impl Config {
    pub fn new<P: AsRef<Path>>(config_file: Option<P>) -> Result<Self, ConfigError> {
        let cfg = load_config(config_file).map_err(ConfigError::ParseError)?;

        // Generate config
        let out: Self = cfg.try_deserialize().map_err(ConfigError::ParseError)?;

        // Validate config
        let context = ValidateContext { config: &out };
        let mut errors = ValidationErrors::new();

        if let Err(e) = out.validate_with_args(&context) {
            errors = e;
        }

        errors.merge_self("profiles", validate_all_with_args(&out.profiles, &context));
        errors.merge_self("machines", validate_all_with_args(&out.machines, &context));
        errors.merge_self(
            "templates",
            validate_all_with_args(&out.templates, &context),
        );
        errors.merge_self("networks", validate_all_with_args(&out.networks, &context));
        errors.merge_self(
            "blueprints",
            validate_all_with_args(&out.blueprints, &context),
        );
        errors.merge_self(
            "allocation",
            validate_all_with_args(&out.allocations, &context),
        );
        errors.merge_self("models", validate_all_with_args(&out.models, &context));

        if !errors.is_empty() {
            return Err(ConfigError::ValidateError(errors));
        }

        Ok(out)
    }
}

fn validate_instances_exist(
    value: &Vec<Instance>,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    for v in value {
        match v {
            Instance::Machine { machine } => validate_machines_exist(machine, context)?,
            Instance::Template { template } => validate_templates_exist(template, context)?,
        }
    }
    Ok(())
}

fn validate_network_exists(value: &str, context: &ValidateContext) -> Result<(), ValidationError> {
    if !context.config.networks.contains_key(value) {
        return Err(ValidationError::new("noexist").with_message("network does not exist".into()));
    }
    Ok(())
}

fn validate_blueprint_exists(
    value: &str,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    if !context.config.blueprints.contains_key(value) {
        return Err(ValidationError::new("noexist").with_message("blueprint does not exist".into()));
    }
    Ok(())
}

fn validate_allocation_exists(
    value: &str,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    if !context.config.allocations.contains_key(value) {
        return Err(
            ValidationError::new("noexist").with_message("allocation does not exist".into())
        );
    }
    Ok(())
}

fn validate_model_exists(value: &str, context: &ValidateContext) -> Result<(), ValidationError> {
    if !context.config.networks.contains_key(value) {
        return Err(ValidationError::new("noexist").with_message("model does not exist".into()));
    }
    Ok(())
}

fn validate_models_exist(
    value: &WeightedList<String>,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    value.try_for_each(|x| validate_model_exists(x, context))
}

fn validate_machine_exists(value: &str, context: &ValidateContext) -> Result<(), ValidationError> {
    if !context.config.machines.contains_key(value) {
        return Err(ValidationError::new("noexist").with_message("machine does not exist".into()));
    }
    Ok(())
}

fn validate_machines_exist(
    value: &WeightedList<String>,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    value.try_for_each(|x| validate_machine_exists(x, context))
}

fn validate_template_exists(value: &str, context: &ValidateContext) -> Result<(), ValidationError> {
    if !context.config.templates.contains_key(value) {
        return Err(ValidationError::new("noexist").with_message("template does not exist".into()));
    }
    Ok(())
}

fn validate_templates_exist(
    value: &WeightedList<String>,
    context: &ValidateContext,
) -> Result<(), ValidationError> {
    value.try_for_each(|x| validate_template_exists(x, context))
}

fn validate_all_with_args<'v_a, T: ValidateArgs<'v_a, Args = &'v_a ValidateContext<'v_a>>>(
    value: &HashMap<String, T>,
    context: &'v_a ValidateContext,
) -> Result<(), ValidationErrors> {
    let mut errors = ValidationErrors::new();
    for (n, x) in value.values().enumerate() {
        errors.merge_self(
            format!("[{n}]").to_static_str(),
            x.validate_with_args(context),
        );
    }
    if !errors.is_empty() {
        return Err(errors);
    }
    Ok(())
}

fn load_default_config() -> Result<::config::Config, ::config::ConfigError> {
    ::config::Config::builder()
        .add_source(::config::File::from_str(
            PREDEFINED_CONFIG,
            ::config::FileFormat::Yaml,
        ))
        .add_source(::config::File::from_str(
            DEFAULT_CONFIG,
            ::config::FileFormat::Yaml,
        ))
        .build()
}

fn load_config<P: AsRef<Path>>(
    opt_config_file: Option<P>,
) -> Result<::config::Config, ::config::ConfigError> {
    let Some(config_file) = opt_config_file else {
        return load_default_config();
    };
    let config_path = config_file.as_ref();
    let Some(config_file_str) = config_path.to_str() else {
        return Err(::config::ConfigError::Message(
            "config file path is not valid UTF-8".to_owned(),
        ));
    };
    let config = ::config::Config::builder()
        .add_source(::config::File::from_str(
            PREDEFINED_CONFIG,
            ::config::FileFormat::Yaml,
        ))
        .add_source(::config::File::new(
            config_file_str,
            ::config::FileFormat::Yaml,
        ))
        .build()?;

    Ok(config)
}
