use super::*;

#[derive(Debug, Clone)]
pub(super) struct MachineRegistryInner {
    unlocked_inner: Arc<MachineRegistryUnlockedInner>,
    srng: StableRng,
    allocated_machines: imbl::HashSet<MachineStateId>,
    profile_state_allocator: StateAllocator<ProfileState>,
    machine_state_allocator: StateAllocator<MachineState>,
    template_state_allocator: StateAllocator<TemplateState>,
    network_state_allocator: StateAllocator<NetworkState>,
    blueprint_state_allocator: StateAllocator<BlueprintState>,
}

impl MachineRegistryInner {
    ///////////////////////////////////////////////////////////
    /// Public Interface

    pub fn new(unlocked_inner: Arc<MachineRegistryUnlockedInner>) -> Self {
        let srng = StableRng::new(unlocked_inner.config.seed.unwrap_or_default());

        MachineRegistryInner {
            unlocked_inner,
            srng,
            allocated_machines: imbl::HashSet::new(),
            profile_state_allocator: StateAllocator::new(),
            machine_state_allocator: StateAllocator::new(),
            template_state_allocator: StateAllocator::new(),
            network_state_allocator: StateAllocator::new(),
            blueprint_state_allocator: StateAllocator::new(),
        }
    }
    pub fn srng(&mut self) -> &mut StableRng {
        &mut self.srng
    }
    pub fn config(&self) -> &config::Config {
        &self.unlocked_inner.config
    }

    pub fn execute_config(&self, cfg: config::Config) -> MachineRegistryResult<()> {
        // Create all networks

        // Create all blueprints

        // Create all templates

        // Create all machines

        Ok(())
    }

    pub fn allocate(&mut self, profile: String) -> MachineRegistryResult<MachineId> {
        // Get profile definition
        let Some(profile_def) = self.unlocked_inner.config.profiles.get(&profile) else {
            return Err(MachineRegistryError::ProfileNotFound);
        };

        // Get current profile state, creating one if we have not yet started executing the profile
        let profile_state_id = self
            .profile_state_allocator
            .get_or_create_by_name(profile, |id, name| {
                ProfileState::new(id, name, profile_def.clone())
            });

        // Get the next instance from the definition
        loop {
            // Move to the next profile instance
            let mut profile_state = self.profile_states().get_state(profile_state_id)?;
            let Some(instance_def) = profile_state.next_instance() else {
                return Err(MachineRegistryError::ProfileComplete);
            };
            self.profile_states_mut().set_state(profile_state);

            let machine_state_id = match instance_def {
                config::Instance::Machine {
                    machine: machine_names,
                } => {
                    // Filter out machines that are already allocated
                    let opt_machine_states_ids = machine_names.try_filter_map(|name| {
                        let Some(machine_state_id) =
                            self.machine_states().get_state_id_by_name(name)
                        else {
                            return Err(MachineRegistryError::MachineNotFound);
                        };
                        if self.allocated_machines.contains(&machine_state_id) {
                            Ok(None)
                        } else {
                            Ok(Some(machine_state_id))
                        }
                    })?;
                    let Some(machine_state_ids) = opt_machine_states_ids else {
                        // All machines in this instance are allocated
                        continue;
                    };

                    // Choose a machine state to activate
                    let machine_state_id = self.srng.weighted_choice(machine_state_ids);

                    // Activate it
                    self.allocated_machines.insert(machine_state_id);

                    machine_state_id
                }
                config::Instance::Template {
                    template: template_names,
                } => {
                    // Filter out templates that are no longer active
                    let opt_template_states = template_names.try_filter_map(|name| {
                        let Some(template_state_id) =
                            self.template_states().get_state_id_by_name(name)
                        else {
                            return Err(MachineRegistryError::TemplateNotFound);
                        };
                        let template_state = self
                            .template_states()
                            .get_state(template_state_id)
                            .expect("must exist");
                        if !template_state.is_active(self) {
                            Ok(None)
                        } else {
                            Ok(Some(template_state))
                        }
                    })?;
                    let Some(template_states) = opt_template_states else {
                        // No templates in this instance are still active
                        continue;
                    };

                    // Chose a template
                    let mut template_state = self.srng.weighted_choice(template_states);

                    // Generate a machine from the template
                    let machine_state_id = template_state.generate(self)?;

                    // Save the updated template
                    self.template_states_mut().set_state(template_state);

                    machine_state_id
                }
            };

            break Ok(machine_state_id.external_id());
        }
    }

    pub fn release(&mut self, machine_id: MachineId) -> MachineRegistryResult<()> {
        let id = StateId::<MachineState>::new(machine_id);
        if self.allocated_machines.contains(&id) {
            // Was a fixed machine, so we leave the machine state so it can
            // be reallocated later
            self.allocated_machines.remove(&id);
        } else {
            // Was a templated machine, so remove the machine state
            let machine_state = self.machine_states().get_state(id)?;
            machine_state.release(self);
            self.machine_states_mut().release_id(id)?;
        }

        Ok(())
    }

    ///////////////////////////////////////////////////////////
    /// Private Implementation

    pub(super) fn profile_states(&self) -> &StateAllocator<ProfileState> {
        &self.profile_state_allocator
    }
    pub(super) fn machine_states(&self) -> &StateAllocator<MachineState> {
        &self.machine_state_allocator
    }
    pub(super) fn template_states(&self) -> &StateAllocator<TemplateState> {
        &self.template_state_allocator
    }
    pub(super) fn network_states(&self) -> &StateAllocator<NetworkState> {
        &self.network_state_allocator
    }
    pub(super) fn blueprint_states(&self) -> &StateAllocator<BlueprintState> {
        &self.blueprint_state_allocator
    }

    pub(super) fn profile_states_mut(&mut self) -> &mut StateAllocator<ProfileState> {
        &mut self.profile_state_allocator
    }
    pub(super) fn machine_states_mut(&mut self) -> &mut StateAllocator<MachineState> {
        &mut self.machine_state_allocator
    }
    pub(super) fn template_states_mut(&mut self) -> &mut StateAllocator<TemplateState> {
        &mut self.template_state_allocator
    }
    pub(super) fn network_states_mut(&mut self) -> &mut StateAllocator<NetworkState> {
        &mut self.network_state_allocator
    }
    pub(super) fn blueprint_states_mut(&mut self) -> &mut StateAllocator<BlueprintState> {
        &mut self.blueprint_state_allocator
    }

    // pub(super) fn get_or_create_machine_state(
    //     &mut self,
    //     opt_name: Option<String>,
    //     params: config::Machine,
    // ) -> MachineRegistryResult<MachineState> {
    //     // Ensure we don't already have this machine created (name must be unique)
    //     if let Some(name) = &opt_name {
    //         if let Some(machine_id) = self.resolve_to_manager_machine.add(name.clone()).get() {
    //             return Ok(self
    //                 .machine_state_by_id
    //                 .get(&machine_id)
    //                 .cloned()
    //                 .expect("must exist"));
    //         }
    //     }

    //     // Allocate a machine id
    //     let machine_id = self.free_machine_ids.pop().unwrap_or_else(|| {
    //         let x = self.next_machine_id;
    //         self.next_machine_id += 1;
    //         x
    //     });

    //     // Create a new machine state
    //     let machine_state = match MachineState::try_new(
    //         self,
    //         machine_id,
    //         MachineStateName::Machine(name.clone()),
    //         machine_def.clone(),
    //     ) {
    //         Ok(v) => v,
    //         Err(e) => {
    //             // Release the machine id
    //             self.free_machine_ids.push(machine_id);
    //             return Err(e);
    //         }
    //     };

    //     // Store the machine state with its unique id
    //     self.machine_state_by_id.insert(machine_id, machine_state);

    //     // Bind the name to the id
    //     self.resolve_to_manager_machine
    //         .resolve(&name, machine_id)
    //         .expect("must resolve");

    //     // Return the state
    //     Ok(self
    //         .machine_state_by_id
    //         .get(&machine_id)
    //         .cloned()
    //         .expect("must exist"))
    // }

    // pub(super) fn get_machine_state_by_id(
    //     &mut self,
    //     machine_id: MachineId,
    // ) -> MachineRegistryResult<&mut MachineState> {
    //     self.machine_state_by_id
    //         .get_mut(&machine_id)
    //         .ok_or_else(|| MachineRegistryError::MachineNotFound)
    // }

    // pub(super) fn get_or_create_network_state(
    //     &mut self,
    //     name: String,
    // ) -> MachineRegistryResult<NetworkState> {

    //     // Get the network def from the config
    //     let Some(network_def) = self.unlocked_inner.config.networks.get(&name).cloned() else {
    //         return Err(MachineRegistryError::NetworkNotFound);
    //     };

    //     // Ensure we don't already have this network created (name must be unique)
    //     if let Some(network_id) = self.resolve_to_manager_network.add(name.clone()).get() {
    //         return Ok(self
    //             .network_state_by_id
    //             .get(&network_id)
    //             .cloned()
    //             .expect("must exist"));
    //     }

    //     // Allocate a network id
    //     let network_id = self.free_network_ids.pop().unwrap_or_else(|| {
    //         let x = self.next_network_id;
    //         self.next_network_id += 1;
    //         x
    //     });

    //     // Create a new network state
    //     let network_state = match NetworkState::try_new(
    //         self,
    //         network_id,
    //         NetworkStateName::Network(name.clone()),
    //         network_def.clone(),
    //     ) {
    //         Ok(v) => v,
    //         Err(e) => {
    //             // Release the network id
    //             self.free_network_ids.push(network_id);
    //             return Err(e);
    //         }
    //     };

    //     // Store the network state with its unique id
    //     self.network_state_by_id.insert(network_id, network_state);

    //     // Bind the name to the id
    //     self.resolve_to_manager_network
    //         .resolve(&name, network_id)
    //         .expect("must resolve");

    //     // Return the unique id
    //     Ok(self
    //         .network_state_by_id
    //         .get(&network_id)
    //         .cloned()
    //         .expect("must exist"))
    // }

    // pub(super) fn get_network_state_by_name(&self, name: &String) -> Option<NetworkState> {
    //     let network_id = self.resolve_to_manager_network.get(name)?;
    //     self.network_state_by_id.get(&network_id).cloned()
    // }

    // pub(super) fn get_network_state_by_id(
    //     &self,
    //     network_id: NetworkId,
    // ) -> MachineRegistryResult<NetworkState> {
    //     self.network_state_by_id
    //         .get(&network_id)
    //         .cloned()
    //         .ok_or_else(|| MachineRegistryError::NetworkNotFound)
    // }

    // pub(super) fn get_or_create_template_state(
    //     &mut self,
    //     name: &String,
    //     template_def: config::Template,
    // ) -> MachineRegistryResult<&mut TemplateState> {
    //     // Ensure we don't already have this template created (name must be unique)
    //     if self.template_state_by_name.contains_key(name) {
    //         return Ok(self
    //             .template_state_by_name
    //             .get_mut(name)
    //             .expect("must exist"));
    //     }

    //     // Create a new template state
    //     let template_state = match TemplateState::try_new(self, name.clone(), template_def.clone())
    //     {
    //         Ok(v) => v,
    //         Err(e) => {
    //             return Err(e);
    //         }
    //     };

    //     // Store the template state with its name
    //     self.template_state_by_name
    //         .insert(name.clone(), template_state);
    //     Ok(self
    //         .template_state_by_name
    //         .get_mut(name)
    //         .expect("must exist"))
    // }

    // pub(super) fn get_or_create_machine_state_from_template(
    //     &mut self,
    //     name: String,
    //     template_def: config::Template,
    // ) -> MachineRegistryResult<MachineState> {
    //     // Make machine def from current template state
    //     let machine_def = {
    //         // Get the active template state
    //         let template_state = self.get_or_create_template_state(&name, template_def)?;
    //         if !template_state.is_active(self)? {
    //             return Err(MachineRegistryError::TemplateComplete);
    //         }

    //         // Pick or instantiate an available network
    //         template_state.generate(self)?

    //         xxx how to pass through per-network limits
    //     };

    //     // Allocate a machine id
    //     let machine_id = self.free_machine_ids.pop().unwrap_or_else(|| {
    //         let x = self.next_machine_id;
    //         self.next_machine_id += 1;
    //         x
    //     });

    //     // Create a new machine state
    //     let machine_state = match MachineState::try_new(
    //         self,
    //         machine_id,
    //         MachineStateName::Template(name.clone()),
    //         machine_def.clone(),
    //     ) {
    //         Ok(v) => v,
    //         Err(e) => {
    //             // Release the machine id
    //             self.free_machine_ids.push(machine_id);
    //             return Err(e);
    //         }
    //     };

    //     // Store the machine state with its unique id
    //     self.machine_state_by_id.insert(machine_id, machine_state);

    //     // Add to machines for this template
    //     {
    //         let template_state = self.get_template_state(&name).expect("must exist");
    //         template_state.machines.insert(machine_id);
    //     }

    //     // Return the unique id
    //     Ok(machine_id)
    // }
    // pub(super) fn get_template_state(&self, name: &String) -> MachineRegistryResult<TemplateState> {
    //     self.template_state_by_name
    //         .get(name)
    //         .cloned()
    //         .ok_or_else(|| MachineRegistryError::TemplateNotFound)
    // }

    // pub(super) fn get_or_create_network_state_from_machine_location(
    //     &mut self,
    //     machine_location: &config::MachineLocation,
    // ) -> MachineRegistryResult<NetworkState> {
    //     match machine_location {
    //         config::MachineLocation::Network {
    //             network: name,
    //             address4: _,
    //             address6: _,
    //         } => {
    //             let network_def = self
    //                 .unlocked_inner
    //                 .config
    //                 .networks
    //                 .get(name)
    //                 .cloned()
    //                 .expect("config validation is broken");
    //             self.get_or_create_network_state(name.clone(), network_def)
    //         }

    //         config::MachineLocation::Blueprint { blueprint: name } => {
    //             let blueprint_def = self
    //                 .unlocked_inner
    //                 .config
    //                 .blueprints
    //                 .get(name)
    //                 .cloned()
    //                 .expect("config validation is broken");
    //             self.get_or_create_network_state_from_blueprint(name.clone(), blueprint_def)
    //         }
    //     }
    // }
    // pub(super) fn get_or_create_network_state_from_template_location(
    //     &mut self,
    //     template_location: &config::TemplateLocation,
    // ) -> MachineRegistryResult<NetworkState> {
    //     match template_location {
    //         config::TemplateLocation::Network { network } => {
    //             let name = self.unlocked_inner.srng.weighted_choice(network);
    //             let network_def = self
    //                 .unlocked_inner
    //                 .config
    //                 .networks
    //                 .get(name)
    //                 .cloned()
    //                 .expect("config validation is broken");
    //             self.get_or_create_network_state(name.clone(), network_def)
    //         }
    //         config::TemplateLocation::Blueprint { blueprint } => {
    //             let name = self.unlocked_inner.srng.weighted_choice(blueprint);
    //             let blueprint_def = self
    //                 .unlocked_inner
    //                 .config
    //                 .blueprints
    //                 .get(name)
    //                 .cloned()
    //                 .expect("config validation is broken");
    //             self.get_or_create_network_state_from_blueprint(name.clone(), blueprint_def)
    //         }
    //     }
    // }

    // pub(super) fn get_blueprint_state(
    //     &self,
    //     name: &String,
    // ) -> MachineRegistryResult<BlueprintState> {
    //     self.blueprint_state_by_name
    //         .get(name)
    //         .cloned()
    //         .ok_or_else(|| MachineRegistryError::BlueprintNotFound)
    // }

    // pub(super) fn choose_allocation_v4(
    //     &mut self,
    //     allocation: config::Allocation,
    //     prefix: u8,
    // ) -> MachineRegistryResult<Ipv4Net> {
    //     // Get allocation subnet candidates
    //     let mut subnet4 = allocation
    //         .subnets
    //         .subnet4
    //         .clone()
    //         .ok_or(MachineRegistryError::NoAllocation)?;

    //     loop {
    //         // Pick a compatible subnet from the allocation
    //         let subnet = self.unlocked_inner.srng.weighted_choice(&subnet4);

    //         // Allocate within the subnet
    //         match self
    //             .address_pool
    //             .add_random_subnet_v4(subnet, prefix)
    //         {
    //             Some(a) => {
    //                 // Got a sub-allocation
    //                 return Ok(a);
    //             }
    //             None => {
    //                 // No sub-allocation left in this subnet,
    //                 // remove the subnet so we can choose again
    //                 let Some(next_subnet4) = subnet4.filter(|x| x == subnet) else {
    //                     // No subnets left
    //                     break;
    //                 };
    //                 subnet4 = next_subnet4;
    //             }
    //         }
    //     }

    //     // No available allocations left
    //     Err(MachineRegistryError::NoAllocation)
    // }

    // pub(super) fn choose_allocation_v6(
    //     &mut self,
    //     allocation: config::Allocation,
    //     prefix: u8,
    // ) -> MachineRegistryResult<Ipv6Net> {
    //     // Get allocation subnet candidates
    //     let mut subnet6 = allocation
    //         .subnets
    //         .subnet6
    //         .clone()
    //         .ok_or(MachineRegistryError::NoAllocation)?;

    //     loop {
    //         // Pick a compatible subnet from the allocation
    //         let subnet = self.unlocked_inner.srng.weighted_choice(&subnet6);

    //         // Allocate within the subnet
    //         match self
    //             .address_pool
    //             .add_random_subnet_v6(subnet, prefix)
    //         {
    //             Some(a) => {
    //                 // Got a sub-allocation
    //                 return Ok(a);
    //             }
    //             None => {
    //                 // No sub-allocation left in this subnet,
    //                 // remove the subnet so we can choose again
    //                 let Some(next_subnet6) = subnet6.filter(|x| x == subnet) else {
    //                     // No subnets left
    //                     break;
    //                 };
    //                 subnet6 = next_subnet6;
    //             }
    //         }
    //     }

    //     // No available allocations left
    //     Err(MachineRegistryError::NoAllocation)
    // }

    // pub(super) fn get_or_create_blueprint_state(
    //     &mut self,
    //     name: &String,
    //     blueprint_def: config::Blueprint,
    // ) -> MachineRegistryResult<BlueprintState> {
    //     // Ensure we don't already have this blueprint created (name must be unique)
    //     if self.blueprint_state_by_name.contains_key(name) {
    //         return Ok(self
    //             .blueprint_state_by_name
    //             .get(name)
    //             .cloned()
    //             .expect("must exist"));
    //     }

    //     // Create a new blueprint state
    //     let blueprint_state =
    //         match BlueprintState::try_new(self, name.clone(), blueprint_def.clone()) {
    //             Ok(v) => v,
    //             Err(e) => {
    //                 return Err(e);
    //             }
    //         };

    //     // Store the blueprint state with its name
    //     self.blueprint_state_by_name
    //         .insert(name.clone(), blueprint_state);
    //     Ok(self
    //         .blueprint_state_by_name
    //         .get(name)
    //         .cloned()
    //         .expect("must exist"))
    // }

    // pub(super) fn get_or_create_network_state_from_blueprint(
    //     &mut self,
    //     name: String,
    //     blueprint_def: config::Blueprint,
    // ) -> MachineRegistryResult<NetworkState> {
    //     // Get the active blueprint state
    //     let blueprint_state = self.get_or_create_blueprint_state(&name, blueprint_def)?;
    //     if !blueprint_state.is_active(self)? {
    //         return Err(MachineRegistryError::BlueprintComplete);
    //     }

    //     // Make network def from current blueprint state
    //     let machine_def = config::Network {
    //         model: self.unlocked_inner.srng.weighted_choice(blueprint_state),
    //         ipv4: todo!(),
    //         ipv6: todo!(),
    //     };

    //     // Allocate a machine id
    //     let machine_id = self.free_machine_ids.pop().unwrap_or_else(|| {
    //         let x = self.next_machine_id;
    //         self.next_machine_id += 1;
    //         x
    //     });

    //     // Create a new machine state
    //     let machine_state = match MachineState::try_new(
    //         self,
    //         MachineStateName::Template(name.clone()),
    //         machine_def.clone(),
    //         machine_id,
    //     ) {
    //         Ok(v) => v,
    //         Err(e) => {
    //             // Release the machine id
    //             self.free_machine_ids.push(machine_id);
    //             return Err(e);
    //         }
    //     };

    //     // Store the machine state with its unique id
    //     self.machine_state_by_id.insert(machine_id, machine_state);

    //     // Return the unique id
    //     Ok(machine_id)
    // }
}
