mod address_pool;
mod machine_registry_inner;
mod state;

use super::*;

use address_pool::*;
use machine_registry_inner::*;
use state::*;

#[derive(Debug)]
struct Machine {}

#[derive(Debug)]
struct MachineRegistryUnlockedInner {
    config: config::Config,
}

#[derive(Debug, Clone)]
pub enum MachineRegistryError {
    InvalidId,
    InvalidName,
    InvalidPrefix,
    AlreadyAttached,
    NotAttached,
    DuplicateName,
    ProfileComplete,
    TemplateComplete,
    NetworkComplete,
    BlueprintComplete,
    ProfileNotFound,
    MachineNotFound,
    NetworkNotFound,
    TemplateNotFound,
    BlueprintNotFound,
    ModelNotFound,
    NoAllocation,
    ResourceInUse,
}

pub type MachineRegistryResult<T> = Result<T, MachineRegistryError>;

#[derive(Debug, Clone)]
pub struct MachineRegistry {
    inner: Arc<Mutex<MachineRegistryInner>>,
    unlocked_inner: Arc<MachineRegistryUnlockedInner>,
}

impl MachineRegistry {
    ///////////////////////////////////////////////////////////
    /// Public Interface
    pub fn new(config: config::Config) -> Self {
        let unlocked_inner = Arc::new(MachineRegistryUnlockedInner { config });
        Self {
            inner: Arc::new(Mutex::new(MachineRegistryInner::new(
                unlocked_inner.clone(),
            ))),
            unlocked_inner,
        }
    }

    pub fn allocate(&self, profile: String) -> MachineRegistryResult<MachineId> {
        let mut inner = self.inner.lock();
        let saved_state = (*inner).clone();
        match inner.allocate(profile) {
            Ok(v) => Ok(v),
            Err(e) => {
                *inner = saved_state;
                Err(e)
            }
        }
    }

    pub fn release(&self, machine_id: MachineId) -> MachineRegistryResult<()> {
        let mut inner = self.inner.lock();
        let saved_state = (*inner).clone();
        match inner.release(machine_id) {
            Ok(v) => Ok(v),
            Err(e) => {
                *inner = saved_state;
                Err(e)
            }
        }
    }
}
