use super::*;

#[derive(Debug)]
struct BlueprintStateImmutable {
    /// The unique id of this blueprint
    id: BlueprintStateId,
    /// The name of this blueprint state
    name: String,
}

#[derive(Debug, Clone)]
pub struct BlueprintStateIpv4Params {
    pub locations: NetworkLocationsList,
    pub prefix: WeightedList<u8>,
    pub gateway: Option<BlueprintStateGatewayParams>,
}

#[derive(Debug, Clone)]
pub struct BlueprintStateIpv6Params {
    pub locations: NetworkLocationsList,
    pub prefix: WeightedList<u8>,
    pub gateway: Option<BlueprintStateGatewayParams>,
}

#[derive(Debug, Clone)]
pub struct BlueprintStateGatewayParams {
    pub translation: WeightedList<config::Translation>,
    pub upnp: Probability,
    pub locations: Option<MachineLocationsList>,
}

#[derive(Debug, Clone)]
struct BlueprintStateIpv4 {
    params: BlueprintStateIpv4Params,
    gateway: Option<BlueprintStateIpv4Gateway>,
}

#[derive(Debug, Clone)]
struct BlueprintStateIpv6 {
    params: BlueprintStateIpv6Params,
    gateway: Option<BlueprintStateIpv6Gateway>,
}

#[derive(Debug, Clone)]
struct BlueprintStateIpv4Gateway {
    params: BlueprintStateGatewayParams,
}

#[derive(Debug, Clone)]
struct BlueprintStateIpv6Gateway {
    params: BlueprintStateGatewayParams,
}

#[derive(Debug, Clone)]
struct BlueprintStateFields {
    limit_network_count: Option<usize>,
    networks: imbl::Vector<NetworkStateId>,
    model: Option<WeightedList<Arc<String>>>,
    ipv4: Option<BlueprintStateIpv4>,
    ipv6: Option<BlueprintStateIpv6>,
}

#[derive(Debug, Clone)]
pub struct BlueprintState {
    immutable: Arc<BlueprintStateImmutable>,
    fields: Arc<BlueprintStateFields>,
}

pub type BlueprintStateId = StateId<BlueprintState>;

impl BlueprintState {
    pub fn new(id: BlueprintStateId, name: String) -> MachineRegistryResult<BlueprintState> {
        Ok(Self {
            immutable: Arc::new(BlueprintStateImmutable { id, name }),
            fields: Arc::new(BlueprintStateFields {
                limit_network_count: None,
                networks: imbl::Vector::new(),
                model: None,
                ipv4: None,
                ipv6: None,
            }),
        })
    }

    pub fn set_limit_network_count(&mut self, limit_network_count: Option<usize>) {
        // Update fields
        self.fields = Arc::new(BlueprintStateFields {
            limit_network_count,
            ..(*self.fields).clone()
        });
    }

    pub fn set_model(&mut self, model: WeightedList<String>) {
        let model = Some(model.map(|x| Arc::new(x.clone())));
        // Update fields
        self.fields = Arc::new(BlueprintStateFields {
            model,
            ..(*self.fields).clone()
        });
    }

    pub fn set_ipv4(
        &mut self,
        params: BlueprintStateIpv4Params,
        gateway_params: Option<BlueprintStateGatewayParams>,
    ) {
        let mut ipv4 = if let Some(ipv4) = self.fields.ipv4.clone() {
            BlueprintStateIpv4 { params, ..ipv4 }
        } else {
            BlueprintStateIpv4 {
                params,
                gateway: None,
            }
        };

        if ipv4.gateway.is_some() {
            if let Some(gateway_params) = gateway_params {
                ipv4.gateway.as_mut().expect("must exist").params = gateway_params;
            } else {
                ipv4.gateway = None;
            }
        } else if let Some(gateway_params) = gateway_params {
            ipv4.gateway = Some(BlueprintStateIpv4Gateway {
                params: gateway_params,
            })
        }

        // Update fields
        self.fields = Arc::new(BlueprintStateFields {
            ipv4: Some(ipv4),
            ..(*self.fields).clone()
        });
    }

    pub fn set_ipv6(
        &mut self,
        params: BlueprintStateIpv6Params,
        gateway_params: Option<BlueprintStateGatewayParams>,
    ) {
        let mut ipv6 = if let Some(ipv6) = self.fields.ipv6.clone() {
            BlueprintStateIpv6 { params, ..ipv6 }
        } else {
            BlueprintStateIpv6 {
                params,
                gateway: None,
            }
        };

        if ipv6.gateway.is_some() {
            if let Some(gateway_params) = gateway_params {
                ipv6.gateway.as_mut().expect("must exist").params = gateway_params;
            } else {
                ipv6.gateway = None;
            }
        } else if let Some(gateway_params) = gateway_params {
            ipv6.gateway = Some(BlueprintStateIpv6Gateway {
                params: gateway_params,
            })
        }

        // Update fields
        self.fields = Arc::new(BlueprintStateFields {
            ipv6: Some(ipv6),
            ..(*self.fields).clone()
        });
    }

    pub fn is_active(&self, machine_registry_inner: &mut MachineRegistryInner) -> bool {
        // Save a backup of the entire state
        let backup = machine_registry_inner.clone();

        // Make a copy of this blueprint state
        let mut current_state = self.clone();

        // See what would happen if we try to generate this blueprint
        let ok = current_state.generate(machine_registry_inner).is_ok();

        // Restore the backup
        *machine_registry_inner = backup;

        // Return if this worked or not
        ok
    }

    fn generate_model_inner(
        &mut self,
        machine_registry_inner: &mut MachineRegistryInner,
        network_state: &mut NetworkState,
    ) -> MachineRegistryResult<()> {
        let model_name = match self.fields.model.as_ref() {
            Some(models) => (**machine_registry_inner.srng().weighted_choice_ref(models)).clone(),
            None => machine_registry_inner.config().default_model.clone(),
        };
        let Some(model) = machine_registry_inner.config().models.get(&model_name) else {
            return Err(MachineRegistryError::ModelNotFound);
        };

        let params = NetworkStateModelParams {
            latency: Some(model.latency.clone()),
            distance: model.distance.clone(),
            loss: model.loss,
        };
        network_state.set_model(params);
        Ok(())
    }

    /// Network filter that ensures we can allocate an ipv4 gateway address on a network
    fn gateway_network_filter_v4(
        &self,
        machine_registry_inner: &MachineRegistryInner,
        network_state_id: NetworkStateId,
    ) -> MachineRegistryResult<bool> {
        // Get the network state
        let network_state = machine_registry_inner
            .network_states()
            .get_state(network_state_id)?;

        // See if we can allocate on this network
        let can_allocate = network_state.can_allocate_address_v4(None);

        Ok(can_allocate)
    }

    /// Network filter that ensures we can allocate an ipv4 gateway address on a network
    fn gateway_network_filter_v6(
        &self,
        machine_registry_inner: &MachineRegistryInner,
        network_state_id: NetworkStateId,
    ) -> MachineRegistryResult<bool> {
        // Get the network state
        let network_state = machine_registry_inner
            .network_states()
            .get_state(network_state_id)?;

        // See if we can allocate on this network
        let can_allocate = network_state.can_allocate_address_v6(None);

        Ok(can_allocate)
    }

    fn generate_ipv4_inner(
        &mut self,
        machine_registry_inner: &mut MachineRegistryInner,
        network_state: &mut NetworkState,
    ) -> MachineRegistryResult<()> {
        network_state.clear_ipv4(machine_registry_inner)?;
        let Some(ipv4) = self.fields.ipv4.as_ref() else {
            return Ok(());
        };

        // Get maximum prefix
        let max_prefix = ipv4
            .params
            .prefix
            .iter()
            .max()
            .copied()
            .expect("must have at least one element");

        // Get addresses for network
        let (subnet, super_net) = match &ipv4.params.locations {
            NetworkLocationsList::Allocations { allocations } => {
                // Get allocations which have subnets that would fit
                // our maximum requested prefix
                let Some(alloc_subnets) = allocations.try_filter_map(|allocation_name| {
                    let allocation = machine_registry_inner
                        .config()
                        .allocations
                        .get(allocation_name)
                        .expect("must exist");
                    Ok(allocation
                        .subnets
                        .subnet4
                        .as_ref()
                        .and_then(|subnet| subnet.filter(|p| p.prefix_len() <= max_prefix)))
                })?
                else {
                    return Err(MachineRegistryError::NoAllocation);
                };

                // Pick an allocation
                let subnets = machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&alloc_subnets);

                // Pick a subnet
                let net = *machine_registry_inner.srng().weighted_choice_ref(subnets);

                // Pick a prefix length that would fit in the subnet
                let opt_subnet = ipv4
                    .params
                    .prefix
                    .filter(|p| *p >= net.prefix_len())
                    .as_ref()
                    .map(|wl| {
                        let subnet_prefix = *machine_registry_inner.srng().weighted_choice_ref(wl);

                        // Use an address pool temporarily to pick a subnet
                        let mut address_pool = AddressPool::<()>::new();
                        address_pool.add_scope_v4(net);
                        address_pool.allocate_random_v4(
                            machine_registry_inner.srng(),
                            subnet_prefix,
                            (),
                        )
                    })
                    .transpose()?
                    .flatten();
                let Some(subnet) = opt_subnet else {
                    return Err(MachineRegistryError::NoAllocation);
                };
                (subnet, None)
            }
            NetworkLocationsList::Networks { networks } => {
                // Get networks which have subnets that would fit
                // our maximum requested prefix
                let Some(available_networks) = networks.try_filter(|network_id| {
                    let super_network_state = machine_registry_inner
                        .network_states()
                        .get_state(*network_id)
                        .expect("must exist");

                    Ok(super_network_state.can_allocate_subnet_v4(None, max_prefix))
                })?
                else {
                    return Err(MachineRegistryError::NoAllocation);
                };

                // Pick a network
                let super_network_id = *machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&available_networks);
                let mut super_network_state = machine_registry_inner
                    .network_states()
                    .get_state(super_network_id)
                    .expect("must exist");

                // Pick a prefix that fits in this network and allocate from it
                let opt_subnet = ipv4
                    .params
                    .prefix
                    .filter(|p| super_network_state.can_allocate_subnet_v4(None, *p))
                    .as_ref()
                    .map(|wl| {
                        let subnet_prefix = *machine_registry_inner.srng().weighted_choice_ref(wl);

                        // Allocate subnet from this network
                        super_network_state.allocate_subnet_v4(
                            machine_registry_inner,
                            OwnerTag::Network(super_network_state.id()),
                            None,
                            subnet_prefix,
                        )
                    })
                    .transpose()?;
                let Some(subnet) = opt_subnet else {
                    return Err(MachineRegistryError::NoAllocation);
                };

                // Update network state
                machine_registry_inner
                    .network_states_mut()
                    .set_state(super_network_state);

                (subnet, Some(super_network_id))
            }
        };

        let params = NetworkStateIpv4Params {
            allocation: subnet,
            super_net,
        };

        let gateway_params = match ipv4.gateway.as_ref() {
            Some(v4gw) => {
                let translation = *machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&v4gw.params.translation);
                let upnp = machine_registry_inner
                    .srng()
                    .probability_test(v4gw.params.upnp);

                let (external_network, external_address) = match v4gw.params.locations.as_ref() {
                    Some(locations_list) => {
                        // A external network location was specified, pick one
                        // Get a network to generate the machine on
                        let mut gateway_network_state = locations_list.pick(
                            machine_registry_inner,
                            |machine_registry_inner, id| {
                                self.gateway_network_filter_v4(machine_registry_inner, id)
                            },
                        )?;
                        let gateway_network_state_id = gateway_network_state.id();

                        // Allocate an external address on this network
                        let external_interface_address = gateway_network_state
                            .allocate_address_v4(
                                machine_registry_inner,
                                OwnerTag::Gateway(network_state.id()),
                                None,
                            )?;

                        // Update the network state
                        machine_registry_inner
                            .network_states_mut()
                            .set_state(gateway_network_state);

                        (
                            gateway_network_state_id,
                            Some(external_interface_address.ip),
                        )
                    }
                    None => {
                        // No external network specified for gateway machine
                        // So use the same network as ourselves
                        (network_state.id(), None)
                    }
                };

                Some(NetworkStateIpv4GatewayParams {
                    translation,
                    upnp,
                    external_network,
                    internal_address: None,
                    external_address,
                })
            }
            None => None,
        };

        network_state.set_ipv4(machine_registry_inner, params, gateway_params)?;
        Ok(())
    }

    fn generate_ipv6_inner(
        &mut self,
        machine_registry_inner: &mut MachineRegistryInner,
        network_state: &mut NetworkState,
    ) -> MachineRegistryResult<()> {
        network_state.clear_ipv6(machine_registry_inner)?;
        let Some(ipv6) = self.fields.ipv6.as_ref() else {
            return Ok(());
        };

        // Get maximum prefix
        let max_prefix = ipv6
            .params
            .prefix
            .iter()
            .max()
            .copied()
            .expect("must have at least one element");

        // Get addresses for network
        let (subnet, super_net) = match &ipv6.params.locations {
            NetworkLocationsList::Allocations { allocations } => {
                // Get allocations which have subnets that would fit
                // our maximum requested prefix
                let Some(alloc_subnets) = allocations.try_filter_map(|allocation_name| {
                    let allocation = machine_registry_inner
                        .config()
                        .allocations
                        .get(allocation_name)
                        .expect("must exist");
                    Ok(allocation
                        .subnets
                        .subnet6
                        .as_ref()
                        .and_then(|subnet| subnet.filter(|p| p.prefix_len() <= max_prefix)))
                })?
                else {
                    return Err(MachineRegistryError::NoAllocation);
                };

                // Pick an allocation
                let subnets = machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&alloc_subnets);

                // Pick a subnet
                let net = *machine_registry_inner.srng().weighted_choice_ref(subnets);

                // Pick a prefix length that would fit in the subnet
                let opt_subnet = ipv6
                    .params
                    .prefix
                    .filter(|p| *p >= net.prefix_len())
                    .as_ref()
                    .map(|wl| {
                        let subnet_prefix = *machine_registry_inner.srng().weighted_choice_ref(wl);

                        // Use an address pool temporarily to pick a subnet
                        let mut address_pool = AddressPool::<()>::new();
                        address_pool.add_scope_v6(net);
                        address_pool.allocate_random_v6(
                            machine_registry_inner.srng(),
                            subnet_prefix,
                            (),
                        )
                    })
                    .transpose()?
                    .flatten();
                let Some(subnet) = opt_subnet else {
                    return Err(MachineRegistryError::NoAllocation);
                };
                (subnet, None)
            }
            NetworkLocationsList::Networks { networks } => {
                // Get networks which have subnets that would fit
                // our maximum requested prefix
                let Some(available_networks) = networks.try_filter(|network_id| {
                    let super_network_state = machine_registry_inner
                        .network_states()
                        .get_state(*network_id)
                        .expect("must exist");

                    Ok(super_network_state.can_allocate_subnet_v6(None, max_prefix))
                })?
                else {
                    return Err(MachineRegistryError::NoAllocation);
                };

                // Pick a network
                let super_network_id = *machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&available_networks);
                let mut super_network_state = machine_registry_inner
                    .network_states()
                    .get_state(super_network_id)
                    .expect("must exist");

                // Pick a prefix that fits in this network and allocate from it
                let opt_subnet = ipv6
                    .params
                    .prefix
                    .filter(|p| super_network_state.can_allocate_subnet_v6(None, *p))
                    .as_ref()
                    .map(|wl| {
                        let subnet_prefix = *machine_registry_inner.srng().weighted_choice_ref(wl);

                        // Allocate subnet from this network
                        super_network_state.allocate_subnet_v6(
                            machine_registry_inner,
                            OwnerTag::Network(super_network_state.id()),
                            None,
                            subnet_prefix,
                        )
                    })
                    .transpose()?;
                let Some(subnet) = opt_subnet else {
                    return Err(MachineRegistryError::NoAllocation);
                };

                // Update network state
                machine_registry_inner
                    .network_states_mut()
                    .set_state(super_network_state);

                (subnet, Some(super_network_id))
            }
        };

        let params = NetworkStateIpv6Params {
            allocation: subnet,
            super_net,
        };

        let gateway_params = match ipv6.gateway.as_ref() {
            Some(v6gw) => {
                let translation = *machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&v6gw.params.translation);
                let upnp = machine_registry_inner
                    .srng()
                    .probability_test(v6gw.params.upnp);

                let (external_network, external_address) = match v6gw.params.locations.as_ref() {
                    Some(locations_list) => {
                        // A external network location was specified, pick one
                        // Get a network to generate the machine on
                        let mut gateway_network_state = locations_list.pick(
                            machine_registry_inner,
                            |machine_registry_inner, id| {
                                self.gateway_network_filter_v6(machine_registry_inner, id)
                            },
                        )?;
                        let gateway_network_state_id = gateway_network_state.id();

                        // Allocate an external address on this network
                        let external_interface_address = gateway_network_state
                            .allocate_address_v6(
                                machine_registry_inner,
                                OwnerTag::Gateway(network_state.id()),
                                None,
                            )?;

                        // Update the network state
                        machine_registry_inner
                            .network_states_mut()
                            .set_state(gateway_network_state);

                        (
                            gateway_network_state_id,
                            Some(external_interface_address.ip),
                        )
                    }
                    None => {
                        // No external network specified for gateway machine
                        // So use the same network as ourselves
                        (network_state.id(), None)
                    }
                };

                Some(NetworkStateIpv6GatewayParams {
                    translation,
                    upnp,
                    external_network,
                    internal_address: None,
                    external_address,
                })
            }
            None => None,
        };

        network_state.set_ipv6(machine_registry_inner, params, gateway_params)?;
        Ok(())
    }

    pub fn generate(
        &mut self,
        machine_registry_inner: &mut MachineRegistryInner,
    ) -> MachineRegistryResult<NetworkStateId> {
        // See if there's room for another network
        if let Some(limit_network_count) = self.fields.limit_network_count {
            if self.fields.networks.len() >= limit_network_count {
                return Err(MachineRegistryError::BlueprintComplete);
            }
        }

        // Allocate a network id
        let network_state_id = machine_registry_inner.network_states_mut().allocate_id();

        // Create an anonymous network state
        let mut network_state =
            NetworkState::new(network_state_id, None, NetworkOrigin::Blueprint(self.id()));

        if let Err(e) = (|| {
            self.generate_model_inner(machine_registry_inner, &mut network_state)?;
            self.generate_ipv4_inner(machine_registry_inner, &mut network_state)?;
            self.generate_ipv6_inner(machine_registry_inner, &mut network_state)?;
            Ok(())
        })() {
            // Release the network state and id if things failed to allocate
            network_state.release(machine_registry_inner);
            machine_registry_inner
                .network_states_mut()
                .release_id(network_state_id)
                .expect("must succeed");
            return Err(e);
        }

        // Attach the state to the id
        machine_registry_inner
            .network_states_mut()
            .attach_state(network_state)?;

        // Record the newly instantiated network
        let mut networks = self.fields.networks.clone();
        networks.push_back(network_state_id);

        // Update fields
        self.fields = Arc::new(BlueprintStateFields {
            networks,
            ..(*self.fields).clone()
        });

        Ok(network_state_id)
    }

    pub fn for_each_network_id<F, R>(&self, mut callback: F) -> MachineRegistryResult<Option<R>>
    where
        F: FnMut(NetworkStateId) -> MachineRegistryResult<Option<R>>,
    {
        for network_id in &self.fields.networks {
            if let Some(res) = callback(*network_id)? {
                return Ok(Some(res));
            }
        }
        Ok(None)
    }

    pub fn on_network_released(&mut self, network_id: NetworkStateId) {
        // Remove network from list
        let pos = self
            .fields
            .networks
            .iter()
            .position(|id| *id == network_id)
            .expect("must exist");
        let mut networks = self.fields.networks.clone();
        networks.remove(pos);

        // Update fields
        self.fields = Arc::new(BlueprintStateFields {
            networks,
            ..(*self.fields).clone()
        });
    }
}

impl State for BlueprintState {
    fn id(&self) -> StateId<Self> {
        self.immutable.id
    }

    fn name(&self) -> Option<String> {
        Some(self.immutable.name.clone())
    }
}
