use super::*;

#[derive(Debug)]
pub enum ResolveToError {
    MissingSymbol,
    AlreadyResolved,
}

pub type ResolveToResult<T> = Result<T, ResolveToError>;

#[derive(Clone, Debug)]
pub struct ResolvesTo<I>
where
    I: Clone + fmt::Debug,
{
    value: Arc<Mutex<Option<I>>>,
}

impl<I> ResolvesTo<I>
where
    I: Clone + fmt::Debug,
{
    pub fn get(&self) -> Option<I> {
        self.value.lock().clone()
    }
}

#[derive(Debug)]
pub struct ResolveToManager<T, I>
where
    T: PartialEq + Eq + PartialOrd + Ord + fmt::Debug,
    I: Clone + fmt::Debug,
{
    symbols: BTreeMap<Arc<T>, Arc<Mutex<Option<I>>>>,
}

impl<T, I> ResolveToManager<T, I>
where
    T: PartialEq + Eq + PartialOrd + Ord + fmt::Debug,
    I: Clone + fmt::Debug,
{
    pub fn new() -> Self {
        Self {
            symbols: BTreeMap::new(),
        }
    }

    pub fn add(&mut self, symbol: T) -> ResolvesTo<I> {
        let symbol = Arc::new(symbol);
        let value = self
            .symbols
            .entry(symbol.clone())
            .or_insert_with(|| Arc::new(Mutex::new(None)))
            .clone();

        ResolvesTo { value }
    }

    pub fn resolve(&mut self, symbol: &T, value: I) -> ResolveToResult<()> {
        match self.symbols.get_mut(symbol) {
            Some(s) => {
                let mut inner = s.lock();
                if inner.is_some() {
                    return Err(ResolveToError::AlreadyResolved);
                }
                *inner = Some(value);
                Ok(())
            }
            None => Err(ResolveToError::MissingSymbol),
        }
    }

    pub fn get(&self, symbol: &T) -> Option<I> {
        self.symbols
            .get(symbol)
            .map(|s| {
                let inner = s.lock();
                inner.clone()
            })
            .flatten()
    }
}
